"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* --------------------------------------------------------------------------------------------
 * Copyright (c) Remy Suen. All rights reserved.
 * Licensed under the MIT License. See License.txt in the project root for license information.
 * ------------------------------------------------------------------------------------------ */
const vscode_languageserver_1 = require("vscode-languageserver");
const dockerfileParser_1 = require("./parser/dockerfileParser");
const docker_1 = require("./docker");
var ValidationCode;
(function (ValidationCode) {
    ValidationCode[ValidationCode["CASING_INSTRUCTION"] = 0] = "CASING_INSTRUCTION";
    ValidationCode[ValidationCode["CASING_DIRECTIVE"] = 1] = "CASING_DIRECTIVE";
    ValidationCode[ValidationCode["ARGUMENT_MISSING"] = 2] = "ARGUMENT_MISSING";
    ValidationCode[ValidationCode["ARGUMENT_EXTRA"] = 3] = "ARGUMENT_EXTRA";
    ValidationCode[ValidationCode["ARGUMENT_REQUIRES_ONE"] = 4] = "ARGUMENT_REQUIRES_ONE";
    ValidationCode[ValidationCode["ARGUMENT_REQUIRES_TWO"] = 5] = "ARGUMENT_REQUIRES_TWO";
    ValidationCode[ValidationCode["ARGUMENT_REQUIRES_ONE_OR_THREE"] = 6] = "ARGUMENT_REQUIRES_ONE_OR_THREE";
    ValidationCode[ValidationCode["NO_SOURCE_IMAGE"] = 7] = "NO_SOURCE_IMAGE";
    ValidationCode[ValidationCode["INVALID_ESCAPE_DIRECTIVE"] = 8] = "INVALID_ESCAPE_DIRECTIVE";
    ValidationCode[ValidationCode["INVALID_AS"] = 9] = "INVALID_AS";
    ValidationCode[ValidationCode["INVALID_PORT"] = 10] = "INVALID_PORT";
    ValidationCode[ValidationCode["INVALID_SIGNAL"] = 11] = "INVALID_SIGNAL";
    ValidationCode[ValidationCode["SYNTAX_MISSING_EQUALS"] = 12] = "SYNTAX_MISSING_EQUALS";
    ValidationCode[ValidationCode["UNKNOWN_INSTRUCTION"] = 13] = "UNKNOWN_INSTRUCTION";
    ValidationCode[ValidationCode["DEPRECATED_MAINTAINER"] = 14] = "DEPRECATED_MAINTAINER";
})(ValidationCode = exports.ValidationCode || (exports.ValidationCode = {}));
var ValidationSeverity;
(function (ValidationSeverity) {
    ValidationSeverity[ValidationSeverity["IGNORE"] = 0] = "IGNORE";
    ValidationSeverity[ValidationSeverity["WARNING"] = 1] = "WARNING";
    ValidationSeverity[ValidationSeverity["ERROR"] = 2] = "ERROR";
})(ValidationSeverity = exports.ValidationSeverity || (exports.ValidationSeverity = {}));
exports.ValidatorSettingsDefaults = {
    deprecatedMaintainer: ValidationSeverity.WARNING,
    directiveCasing: ValidationSeverity.WARNING,
    instructionCasing: ValidationSeverity.WARNING
};
class Validator {
    constructor(settings) {
        if (!settings) {
            settings = exports.ValidatorSettingsDefaults;
        }
        this.settings = settings;
    }
    parseDirective(dockerfile, document, problems) {
        let directive = dockerfile.getDirective();
        if (directive === null) {
            return;
        }
        let directiveName = directive.getDirective();
        let value = directive.getValue();
        if (directiveName === docker_1.DIRECTIVE_ESCAPE) {
            if (value !== '\\' && value !== '`' && value !== "") {
                // if the directive's value is invalid or isn't the empty string, flag it
                let range = directive.getValueRange();
                problems.push(Validator.createInvalidEscapeDirective(range.start, range.end, value));
            }
            if (directive.getName() !== docker_1.DIRECTIVE_ESCAPE) {
                let range = directive.getNameRange();
                let diagnostic = this.createLowercaseDirective(range.start, range.end);
                if (diagnostic) {
                    problems.push(diagnostic);
                }
            }
        }
    }
    /**
     * Checks the arguments of the given instruction.
     *
     * @param instruction the instruction to validate
     * @param problems an array of identified problems in the document
     * @param expectedArgCount an array of expected number of arguments
     *                         for the instruction, if its length is 1
     *                         and its value is -1, any number of
     *                         arguments greater than zero is valid
     * @param validate the function to use to validate an argument
     * @param createIncompleteDiagnostic the function to use to create a diagnostic
     *                                   if the number of arguments is incorrect
     */
    checkArguments(instruction, problems, expectedArgCount, validate, createIncompleteDiagnostic) {
        let args = instruction.getArguments();
        if (args.length === 0) {
            // all instructions are expected to have at least one argument
            let range = instruction.getInstructionRange();
            problems.push(Validator.createMissingArgument(range.start, range.end));
        }
        else if (expectedArgCount[0] === -1) {
            for (let i = 0; i < args.length; i++) {
                let createInvalidDiagnostic = validate(i, args[i].getValue());
                if (createInvalidDiagnostic) {
                    let range = args[i].getRange();
                    problems.push(createInvalidDiagnostic(range.start, range.end, args[i].getValue()));
                }
            }
        }
        else {
            for (let i = 0; i < expectedArgCount.length; i++) {
                if (expectedArgCount[i] === args.length) {
                    for (let j = 0; j < args.length; j++) {
                        let createInvalidDiagnostic = validate(j, args[j].getValue());
                        if (createInvalidDiagnostic) {
                            let range = args[j].getRange();
                            problems.push(createInvalidDiagnostic(range.start, range.end, args[i].getValue()));
                        }
                    }
                    return;
                }
            }
            let range = args[args.length - 1].getRange();
            if (createIncompleteDiagnostic) {
                problems.push(createIncompleteDiagnostic(range.start, range.end));
            }
            else {
                problems.push(Validator.createExtraArgument(range.start, range.end));
            }
        }
    }
    validate(keywords, document) {
        this.document = document;
        let problems = [];
        let parser = new dockerfileParser_1.DockerfileParser();
        let dockerfile = parser.parse(document);
        this.parseDirective(dockerfile, document, problems);
        let instructions = dockerfile.getInstructions();
        if (instructions.length === 0 || dockerfile.getARGs().length === instructions.length) {
            // no instructions in this file, or only ARGs
            problems.push(Validator.createNoSourceImage(document.positionAt(0), document.positionAt(0)));
        }
        let hasFrom = false;
        for (let instruction of dockerfile.getInstructions()) {
            let keyword = instruction.getKeyword();
            if (keyword === "FROM") {
                hasFrom = true;
            }
            else if (!hasFrom && keyword !== "ARG") {
                // first non-ARG instruction is not a FROM
                let range = instruction.getInstructionRange();
                problems.push(Validator.createNoSourceImage(range.start, range.end));
                hasFrom = true;
            }
            if (keywords.indexOf(keyword) === -1) {
                let range = instruction.getInstructionRange();
                // invalid instruction found
                problems.push(Validator.createUnknownInstruction(range.start, range.end, keyword));
            }
            else {
                if (keyword !== instruction.getInstruction()) {
                    let range = instruction.getInstructionRange();
                    // warn about uppercase convention if the keyword doesn't match the actual instruction
                    let diagnostic = this.createUppercaseInstruction(range.start, range.end);
                    if (diagnostic) {
                        problems.push(diagnostic);
                    }
                }
                if (keyword === "MAINTAINER") {
                    let range = instruction.getInstructionRange();
                    let diagnostic = this.createMaintainerDeprecated(range.start, range.end);
                    if (diagnostic) {
                        problems.push(diagnostic);
                    }
                }
                switch (keyword) {
                    case "ARG":
                        this.checkArguments(instruction, problems, [-1], function (index, argument) {
                            if (index > 0) {
                                return Validator.createARGRequiresOneArgument;
                            }
                            return null;
                        }, Validator.createARGRequiresOneArgument);
                        break;
                    case "ENV":
                        this.checkArguments(instruction, problems, [-1], function () {
                            return null;
                        });
                        let env = instruction;
                        let properties = env.getProperties();
                        if (properties.length === 1 && properties[0].getValue() === null) {
                            let range = properties[0].getNameRange();
                            problems.push(Validator.createENVRequiresTwoArguments(range.start, range.end));
                        }
                        else if (properties.length !== 0) {
                            for (let property of properties) {
                                if (property.getValue() === null) {
                                    let range = property.getNameRange();
                                    problems.push(Validator.createSyntaxMissingEquals(range.start, range.end, property.getName()));
                                }
                            }
                        }
                        break;
                    case "FROM":
                        this.checkArguments(instruction, problems, [1, 3], function (index, argument) {
                            if (index === 1) {
                                return argument.toUpperCase() === "AS" ? null : Validator.createInvalidAs;
                            }
                            return null;
                        }, Validator.createRequiresOneOrThreeArguments);
                        break;
                    case "STOPSIGNAL":
                        this.checkArguments(instruction, problems, [1], function (index, argument) {
                            if (argument.indexOf("SIG") === 0 || argument.indexOf('$') != -1) {
                                return null;
                            }
                            for (var i = 0; i < argument.length; i++) {
                                if ('0' > argument.charAt(i) || '9' < argument.charAt(i)) {
                                    return Validator.createInvalidStopSignal;
                                }
                            }
                            return null;
                        });
                        break;
                    case "EXPOSE":
                        this.checkArguments(instruction, problems, [-1], function (index, argument) {
                            for (var i = 0; i < argument.length; i++) {
                                if (argument.charAt(i) !== '-' && ('0' > argument.charAt(i) || '9' < argument.charAt(i))) {
                                    return Validator.createInvalidPort;
                                }
                            }
                            return argument.charAt(0) !== '-' && argument.charAt(argument.length - 1) !== '-' ? null : Validator.createInvalidPort;
                        });
                        break;
                    default:
                        this.checkArguments(instruction, problems, [-1], function () {
                            return null;
                        });
                        break;
                }
            }
        }
        return problems;
    }
    static formatMessage(text, variable) {
        return text.replace("${0}", variable);
    }
    static getDiagnosticMessage_DirectiveCasing() {
        return Validator.dockerProblems["directiveCasing"];
    }
    static getDiagnosticMessage_DirectiveEscapeInvalid(value) {
        return Validator.formatMessage(Validator.dockerProblems["directiveEscapeInvalid"], value);
    }
    static getDiagnosticMessage_NoSourceImage() {
        return Validator.dockerProblems["noSourceImage"];
    }
    static getDiagnosticMessage_InvalidAs() {
        return Validator.dockerProblems["invalidAs"];
    }
    static getDiagnosticMessage_InvalidPort(port) {
        return Validator.formatMessage(Validator.dockerProblems["invalidPort"], port);
    }
    static getDiagnosticMessage_InvalidSignal(signal) {
        return Validator.formatMessage(Validator.dockerProblems["invalidStopSignal"], signal);
    }
    static getDiagnosticMessage_InstructionExtraArgument() {
        return Validator.dockerProblems["instructionExtraArgument"];
    }
    static getDiagnosticMessage_InstructionMissingArgument() {
        return Validator.dockerProblems["instructionMissingArgument"];
    }
    static getDiagnosticMessage_ARGRequiresOneArgument() {
        return Validator.formatMessage(Validator.dockerProblems["instructionRequiresOneArgument"], "ARG");
    }
    static getDiagnosticMessage_ENVRequiresTwoArguments() {
        return Validator.formatMessage(Validator.dockerProblems["instructionRequiresTwoArguments"], "ENV");
    }
    static getDiagnosticMessage_InstructionRequiresOneOrThreeArguments() {
        return Validator.dockerProblems["fromRequiresOneOrThreeArguments"];
    }
    static getDiagnosticMessage_InstructionUnknown(instruction) {
        return Validator.formatMessage(Validator.dockerProblems["instructionUnknown"], instruction);
    }
    static getDiagnosticMessage_SyntaxMissingEquals(argument) {
        return Validator.formatMessage(Validator.dockerProblems["syntaxMissingEquals"], argument);
    }
    static getDiagnosticMessage_InstructionCasing() {
        return Validator.dockerProblems["instructionCasing"];
    }
    static getDiagnosticMessage_DeprecatedMaintainer() {
        return Validator.dockerProblems["deprecatedMaintainer"];
    }
    static createInvalidEscapeDirective(start, end, value) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_DirectiveEscapeInvalid(value), ValidationCode.INVALID_ESCAPE_DIRECTIVE);
    }
    static createInvalidAs(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InvalidAs(), ValidationCode.INVALID_AS);
    }
    static createInvalidPort(start, end, port) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InvalidPort(port), ValidationCode.INVALID_PORT);
    }
    static createInvalidStopSignal(start, end, signal) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InvalidSignal(signal), ValidationCode.INVALID_SIGNAL);
    }
    static createMissingArgument(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InstructionMissingArgument(), ValidationCode.ARGUMENT_MISSING);
    }
    static createExtraArgument(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InstructionExtraArgument(), ValidationCode.ARGUMENT_EXTRA);
    }
    static createARGRequiresOneArgument(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_ARGRequiresOneArgument(), ValidationCode.ARGUMENT_REQUIRES_ONE);
    }
    static createENVRequiresTwoArguments(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_ENVRequiresTwoArguments(), ValidationCode.ARGUMENT_REQUIRES_TWO);
    }
    static createRequiresOneOrThreeArguments(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InstructionRequiresOneOrThreeArguments(), ValidationCode.ARGUMENT_REQUIRES_ONE_OR_THREE);
    }
    static createNoSourceImage(start, end) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_NoSourceImage(), ValidationCode.NO_SOURCE_IMAGE);
    }
    static createSyntaxMissingEquals(start, end, argument) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_SyntaxMissingEquals(argument), ValidationCode.SYNTAX_MISSING_EQUALS);
    }
    static createUnknownInstruction(start, end, instruction) {
        return Validator.createError(start, end, Validator.getDiagnosticMessage_InstructionUnknown(instruction), ValidationCode.UNKNOWN_INSTRUCTION);
    }
    static createError(start, end, description, code) {
        return Validator.createDiagnostic(vscode_languageserver_1.DiagnosticSeverity.Error, start, end, description, code);
    }
    createLowercaseDirective(start, end) {
        if (this.settings.directiveCasing === ValidationSeverity.ERROR) {
            return Validator.createError(start, end, Validator.getDiagnosticMessage_DirectiveCasing(), ValidationCode.CASING_DIRECTIVE);
        }
        else if (this.settings.directiveCasing === ValidationSeverity.WARNING) {
            return Validator.createWarning(start, end, Validator.getDiagnosticMessage_DirectiveCasing(), ValidationCode.CASING_DIRECTIVE);
        }
        return null;
    }
    createUppercaseInstruction(start, end) {
        if (this.settings.instructionCasing === ValidationSeverity.ERROR) {
            return Validator.createError(start, end, Validator.getDiagnosticMessage_InstructionCasing(), ValidationCode.CASING_INSTRUCTION);
        }
        else if (this.settings.instructionCasing === ValidationSeverity.WARNING) {
            return Validator.createWarning(start, end, Validator.getDiagnosticMessage_InstructionCasing(), ValidationCode.CASING_INSTRUCTION);
        }
        return null;
    }
    createMaintainerDeprecated(start, end) {
        if (this.settings.deprecatedMaintainer === ValidationSeverity.ERROR) {
            return Validator.createError(start, end, Validator.getDiagnosticMessage_DeprecatedMaintainer(), ValidationCode.DEPRECATED_MAINTAINER);
        }
        else if (this.settings.deprecatedMaintainer === ValidationSeverity.WARNING) {
            return Validator.createWarning(start, end, Validator.getDiagnosticMessage_DeprecatedMaintainer(), ValidationCode.DEPRECATED_MAINTAINER);
        }
        return null;
    }
    static createWarning(start, end, description, code) {
        return Validator.createDiagnostic(vscode_languageserver_1.DiagnosticSeverity.Warning, start, end, description, code);
    }
    static createDiagnostic(severity, start, end, description, code) {
        return {
            range: {
                start: start,
                end: end
            },
            message: description,
            severity: severity,
            code: code,
            source: "dockerfile-lsp"
        };
    }
}
Validator.dockerProblems = {
    "directiveCasing": "Parser directives should be written in lowercase letters",
    "directiveEscapeInvalid": "invalid ESCAPE '${0}'. Must be ` or \\",
    "noSourceImage": "No source image provided with `FROM`",
    "fromRequiresOneOrThreeArguments": "FROM requires either one or three arguments",
    "invalidAs": "Second argument should be AS",
    "invalidPort": "Invalid containerPort: ${0}",
    "invalidStopSignal": "Invalid signal: ${0}",
    "syntaxMissingEquals": "Syntax error - can't find = in \"${0}\". Must be of the form: name=value",
    "instructionExtraArgument": "Instruction has an extra argument",
    "instructionMissingArgument": "Instruction has no arguments",
    "instructionRequiresOneArgument": "${0} requires exactly one argument",
    "instructionRequiresTwoArguments": "${0} must have two arguments",
    "instructionUnknown": "Unknown instruction: ${0}",
    "instructionCasing": "Instructions should be written in uppercase letters",
    "deprecatedMaintainer": "MAINTAINER has been deprecated",
};
exports.Validator = Validator;
